void clock_readtime_monotonic(long* t);
void clock_readtime_realtime(long* t);
void clock_readtime_processtime(long* t);
void clock_readtime_threadtime(long* t);

void clock_readres_monotonic(long* t);
void clock_readres_realtime(long* t);
void clock_readres_processtime(long* t);
void clock_readres_threadtime(long* t);
